// freecd.cpp
//
// FreeCD v1.91 for Win95 / NT 4.0
// by Nate Smith
//
// e-mail: nate@cryogen.com
//
// This source code is PUBLIC DOMAIN:
// You are free to modify and distribute it as you see fit, however,
// I would appreciate being notified of any modifications to avoid
// distribution conflicts.  Thanks.
//
// Please see the README.TXT file that accompanied this source code
// for further information.


////////////////////////////////// Headers ////////////////////////////////////

#define STRICT // strict type-checking
#define WIN32_LEAN_AND_MEAN // only include necessary headers

#include <windows.h>
#include <shellapi.h>
#include <mmsystem.h>
#include <stdlib.h>
#include <string.h>

#include "freecd.h"


//////////////////////////////// Declarations ////////////////////////////////

// declare & initialize global variables
HINSTANCE hInst;
HWND hWndMain,				// main, help, & options window handles;
   hWndHlp = NULL,		// initialize these 2 as NULL so we can
   hWndOptions = NULL;	// tell whether they already exist later...
HICON hIconPlay,
	hIconPause,
   hIconX;
HDC hdcMemMain; // DC handle for window refresh image in memory
HMENU hMenuMain,
	hMenuSys;
enum StatusType { open, playing, paused } Status = open; // current status
enum DispType { norm, remtrk, remdisc, total } Disp = norm; // current time-display mode
char AppName[10] = "FreeCD", // name for window and window-class
   Pos[12] = "01:00:00:00", // disc-position string passed to OutChars()
	SPos[32] = "1  0:00  [FreeCD]"; // disc position string used in button text and taskbar tooltip
POINT newPt,	// current mouse coordinates
	oldPt,		// previous mouse coordinates
   ptTrack,		// reference coordinate for track number bitmaps (see freecdrefs.c)
   ptTime;		// reference coordinate for time bitmaps (see freecdrefs.c)
RECT rcDisp;	// time-display rectangle (see freecdrefs.c); checked against mouse coordinates for dragging the window
BOOL BtnDn = FALSE, // is the mouse button currently pressed; used for dragging the window
	TopMost = FALSE, // always-on-top state
   DiscPres = TRUE, // is a disc currently in the drive
   Repeat = FALSE, // is track-repeat enabled
   Hidden = FALSE, // is the window minimized to the taskbar
   PrevTime = TRUE; // was the time shown in the taskbar button before the options dialog was opened; used when option changes are cancelled
int WinWidth,
	WinHeight,
   WinX,
   WinY,
   Track = 1, // current track
   RptTrack = 0, // the track that was current when repeat was enabled
   NumTracks, // total number of tracks on the disc
   TimeOut = 1000, // delay in milliseconds between calls to TimerProc()
   Wait = 0; // number of <TimeOut>'s to wait until allowing TimerProc() to proceed
UINT wTimerRes, // TimerProc()'s timer resolution
	wTimerID; // TimerProc()'s timer identifier

struct RefreshItems { // used to organize info related to each bitmap in the main window
	int x, y, w, h;
   char bmpup[11], bmpdn[11];
} ref[11];

struct TrackInfo { // used to organize info related to each track on the current disc
	int m, s;
   char str[6];
} trk[99], // info for each track on the disc
	cur, // info for current position in normal mode
   cur2, // info for current position in total-disc-length-remaining mode
   rem, // info for current position in track-length-remaining mode
   disc; // info for total-disc-length info

struct OptionsStruct { // organize the current states of all the options
	BOOL TBIcon,
   	  ScrnPos,
        AutoPlay,
        AutoRpt,
        KeepRpt,
        TaskTime;
} Opt = { FALSE, TRUE, TRUE, TRUE, TRUE, TRUE }; // defaults for the options


/////////////////////////////// Function Prototypes //////////////////////////////
//
// see the function definitions below for details of each one

int PASCAL				WinMain(HINSTANCE, HINSTANCE, LPSTR, int);
void						UnInstall();
void						InitCD(HINSTANCE, HINSTANCE, LPSTR, int);
BOOL						TaskBarAddIcon(LPSTR);
BOOL						TaskBarModifyIcon(LPSTR);
BOOL						TaskBarDeleteIcon();
void						PrimeBmpOps();
void						DrawBmp(int, int, LPSTR);
void						OutChars(char[12]);
BOOL						RevSkip(BOOL);
BOOL						PlayDisc();
void						PauseDisc();
BOOL						FFSkip();
void						GetDiscInfo();
void						TimeToInt(TrackInfo &tr);
void						KillApp();
void						On_CDWM_NOTIFYICON(LPARAM);
BOOL WINAPI				OnContextMenu(int, int);
void APIENTRY			DisplayContextMenu(POINT);
void APIENTRY			DisplayTaskBarMenu();
void						UpdatePlayBtn();
void						OwnerDrawBmp(LPDRAWITEMSTRUCT);
void						DrawBtn(LPDRAWITEMSTRUCT, int);
BOOL CALLBACK			HelpDlgProc(HWND, UINT, WPARAM, LPARAM);
BOOL CALLBACK			OptDlgProc(HWND, UINT, WPARAM, LPARAM);
void FAR PASCAL		TimerProc(UINT wTimerID, UINT msg, DWORD dwUser, DWORD dw1, DWORD dw2);
LRESULT CALLBACK		MainWndProc(HWND, UINT, WPARAM, LPARAM);


////////////////////////// Initialization functions //////////////////////////////

// WinMain
int PASCAL WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
						 LPSTR lpCmdLine, int CmdShow)
{
	MSG msg;

   if (strstr(lpCmdLine, "/u"))	// if FreeCD.exe was called with the "/u" command-line switch,
   	UnInstall();					// halt further execution of the program and uninstall the program
   else									// or if not,
		InitCD(hInstance, hPrevInstance, lpCmdLine, CmdShow); // proceed to run the program

	// standard message relaying loop;
   // first checks for messages directed to either dialog box
   while (GetMessage(&msg, NULL, NULL, NULL)) {
   	if (!IsDialogMessage(hWndHlp, &msg) && !IsDialogMessage(hWndOptions, &msg)) {
      	TranslateMessage(&msg);
      	DispatchMessage(&msg);
   	}
	}
	return (msg.wParam);
}

// -- uninstalls FreeCD --
// this is called if FreeCD.exe is called with the "/u"
// command-line switch, which is typically only used by
// the "Add/Remove Programs" control panel applet;
// NOTE: the "Add/Remove" functionality is added every
// time FreeCD starts up by InitCD()'s registry calls
void UnInstall()
{
	HWND hWndPrev;
   HKEY hKeyFreeCD;
	char buffer[2048];
   char WinInitFile[MAX_PATH];
	char CurrentPath[MAX_PATH];
   char linetoadd[MAX_PATH];
	int length;
   DWORD dwData = (DWORD)MAX_PATH;
   OSVERSIONINFO osvi;

   // see if the user REALLY wants to do this...
   if (MessageBox(NULL, "Are you sure you want to uninstall FreeCD "
   	"and remove it from your system?", "UnInstall?",
      MB_ICONQUESTION | MB_YESNO) == IDYES)
   { // if they're REALLY sure...
		do { // close any running instances
			hWndPrev = FindWindow((LPCTSTR)AppName, NULL);
			SendMessage(hWndPrev, WM_COMMAND, IDBTN_EXIT, NULL);
		} while (hWndPrev);

		// get the most recent path to the file FreeCD.exe;
      // this is saved in the FreeCD reg key every time FreeCD starts up
      RegOpenKeyEx(HKEY_CURRENT_USER, "Software\\Ion\\FreeCD", 0, KEY_READ, &hKeyFreeCD);
      RegQueryValueEx(hKeyFreeCD, "Path", 0, NULL, (LPBYTE)&CurrentPath, (LPDWORD)&dwData);
      RegCloseKey(hKeyFreeCD);

		// remove all the registry keys and values associated with FreeCD
      RegDeleteKey(HKEY_CURRENT_USER, "Software\\Ion\\FreeCD");
	   RegDeleteKey(HKEY_CURRENT_USER, "Software\\Ion");
	   RegDeleteKey(HKEY_LOCAL_MACHINE, "Software\\Microsoft\\Windows\\CurrentVersion\\Uninstall\\FreeCD");

		// find out what operating system is running
		osvi.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
		GetVersionEx(&osvi);

      // if the operating system is WinNT
      if (osvi.dwPlatformId == VER_PLATFORM_WIN32_NT)
      {
			wsprintf(linetoadd, "%s\\FreeCD.exe", CurrentPath);
        	MoveFileEx(linetoadd, NULL, MOVEFILE_REPLACE_EXISTING | MOVEFILE_DELAY_UNTIL_REBOOT);
      }
      else // or, if the operating system is Win95 or Win32s
      {
			GetWindowsDirectory(WinInitFile, MAX_PATH);
			lstrcat(WinInitFile, "\\WININIT.INI");
			length = GetPrivateProfileSection( "Rename", buffer, 2048, WinInitFile );
			wsprintf(linetoadd, "NUL=%s\\FreeCD.exe", CurrentPath);
			// you can't just use lstrcat here because the indvidual lines are separated by \0
			lstrcpy( buffer + length, linetoadd );
			*(buffer + length + lstrlen(linetoadd) + 1) = '\0';  // have to add trailing \0
			WritePrivateProfileSection( "Rename", buffer, WinInitFile );
   	}

      // let them know everything worked as planned
	   MessageBox(NULL, "FreeCD v1.91 has been successfully uninstalled.  The changes will "
	   	"take effect the next time you start Windows.", "UnInstallation Complete",
   	   MB_ICONINFORMATION | MB_OK);
   }
   PostQuitMessage(0); // exit the message loop and close FreeCD
}

// -- perform all preliminary tasks --
// defines the window class, creates the window,
// and retrieves settings from the registry
#pragma argsused
void InitCD(HINSTANCE hInstance, HINSTANCE hPrevInstance,
				LPSTR lpCmdLine, int CmdShow)
{
   HKEY hKeyFreeCD, hKeyUnInst;
   DWORD dwResult;
   char cBuf[9] = "0000000\n";
   char CurrentPath[MAX_PATH];
   DWORD dwData = (DWORD)9;

	// if another instance of FreeCD is running, end this one
	if (FindWindow((LPCTSTR)AppName, NULL)){
		PostQuitMessage(0);
	}
	else{ // or, if this is the first instance...
		WNDCLASSEX wcCDClass;

		wcCDClass.cbSize			= sizeof(WNDCLASSEX);
		wcCDClass.lpszClassName = AppName;
		wcCDClass.hInstance     = hInstance;
		wcCDClass.lpfnWndProc   = MainWndProc;
		wcCDClass.hCursor       = LoadCursor(NULL, IDC_ARROW);
		wcCDClass.hIcon         = LoadIcon(hInstance, "Face");
		wcCDClass.hIconSm			= NULL;
		wcCDClass.lpszMenuName  = NULL;
		wcCDClass.hbrBackground = HBRUSH(COLOR_WINDOWFRAME + 1); // black background
		wcCDClass.style         = CS_DBLCLKS | CS_HREDRAW | CS_VREDRAW; // receive double-click messages
		wcCDClass.cbClsExtra    = 0;
		wcCDClass.cbWndExtra    = 0;

		RegisterClassEx(&wcCDClass);

		// define positions and names of all the button bitmaps
      #include "freecdrefs.c"

		RECT rc;

		hInst = hInstance;

		GetWindowRect(GetDesktopWindow(), &rc); // get the size of the screen

      // the window's top-left corner defaults to the bottom-right corner
      // of the screen; this can be changed to suit your personal preference;
      // remember, the window's position will be adjusted before it becomes
      // visible (see the WM_SHOWWINDOW case in MainWndProc())
		WinX = rc.right;
		WinY = rc.bottom;

		hWndMain = CreateWindow(AppName, AppName,
									WS_POPUP | WS_SYSMENU | WS_DLGFRAME,
									WinX, WinY, WinWidth, WinHeight,
									NULL, NULL, hInstance, NULL);

		// if the FreeCD registry keys don't already exist, create them;
      // otherwise, open the existing keys
		RegCreateKeyEx(HKEY_LOCAL_MACHINE, "Software\\Microsoft\\Windows\\CurrentVersion\\Uninstall\\FreeCD", 0,
      					NULL, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL,
                     &hKeyUnInst, &dwResult);
		RegCreateKeyEx(HKEY_CURRENT_USER, "Software\\Ion\\FreeCD", 0,
      					NULL, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL,
                     &hKeyFreeCD, &dwResult);

		// get the current path to the file FreeCD.exe
      // and save it to the registry for unistallation
	   GetCurrentDirectory(MAX_PATH, CurrentPath);
		RegSetValueEx(hKeyFreeCD, "Path", 0, REG_SZ, (LPBYTE)CurrentPath, strlen(CurrentPath) + 1);

		// set the uninstallation info for the "Add/Remove
      // Programs" control panel applet
		strcat(CurrentPath, "\\FreeCD.exe /u");
		RegSetValueEx(hKeyUnInst, "DisplayName", 0, REG_SZ, (LPBYTE)"FreeCD v1.91", 13);
		RegSetValueEx(hKeyUnInst, "UninstallString", 0, REG_SZ, (LPBYTE)CurrentPath, strlen(CurrentPath) + 1);
      RegCloseKey(hKeyUnInst);

		// don't try to read the previous settings unless the
      // key already existed
		if (dwResult == REG_OPENED_EXISTING_KEY)
      {
         // get each value and copy it into its respective variable
			RegQueryValueEx(hKeyFreeCD, "UseLastWindowPos", 0, NULL, (LPBYTE)&cBuf, (LPDWORD)&dwData);
         dwData = (DWORD)9; // this MUST be reset before each call to RegQueryValueEx()
         Opt.ScrnPos = (strncmp(cBuf, "Ye", 2) ? FALSE : TRUE); // set the option according to the last saved setting
         if (Opt.ScrnPos)
         { // only get the last saved screen position if it is needed
	  		  	RegQueryValueEx(hKeyFreeCD, "WindowPosX", 0, NULL, (LPBYTE)&cBuf, (LPDWORD)&dwData);
            dwData = (DWORD)9;
				WinX = atoi(cBuf);
				RegQueryValueEx(hKeyFreeCD, "WindowPosY", 0, NULL, (LPBYTE)&cBuf, (LPDWORD)&dwData);
            dwData = (DWORD)9;
         	WinY = atoi(cBuf);
            // move the window before it is checked by the WM_SHOWWINDOW case in MainWndProc()
      		SetWindowPos(hWndMain, NULL, WinX, WinY, NULL, NULL,
   	       				 SWP_NOSIZE | SWP_NOZORDER);
         }
			RegQueryValueEx(hKeyFreeCD, "AlwaysMinimized", 0, NULL, (LPBYTE)&cBuf, (LPDWORD)&dwData);
         dwData = (DWORD)9;
         Opt.TBIcon = (strncmp(cBuf, "Ye", 2) ? FALSE : TRUE);
         if (Opt.TBIcon)
         	PostMessage(hWndMain, WM_COMMAND, IDBTN_MIN, NULL); // simulate pressing the minimize button
         else
         { // only read this if it's needed
				RegQueryValueEx(hKeyFreeCD, "Minimized", 0, NULL, (LPBYTE)&cBuf, (LPDWORD)&dwData);
            dwData = (DWORD)9;
      	   if (!strncmp(cBuf, "Ye", 2))
         		PostMessage(hWndMain, WM_COMMAND, IDBTN_MIN, NULL); // simulate pressing the minimize button
   	      else // if it wasn't minimized last time, just show it as usual...
      	   	ShowWindow(hWndMain, SW_SHOW);
         }
		   RegQueryValueEx(hKeyFreeCD, "AlwaysOnTop", 0, NULL, (LPBYTE)&cBuf, (LPDWORD)&dwData);
         dwData = (DWORD)9;
         if (!strncmp(cBuf, "Ye", 2))
         	PostMessage(hWndMain, WM_COMMAND, IDBTN_ONTOP, NULL); // simulate activating always-on-top
         else
         	TopMost = FALSE;
			// the rest of these are saved for later use...
			RegQueryValueEx(hKeyFreeCD, "AutoPlay", 0, NULL, (LPBYTE)&cBuf, (LPDWORD)&dwData);
         dwData = (DWORD)9;
         Opt.AutoPlay = (strncmp(cBuf, "Ye", 2) ? FALSE : TRUE);
			RegQueryValueEx(hKeyFreeCD, "AutoRepeat", 0, NULL, (LPBYTE)&cBuf, (LPDWORD)&dwData);
         dwData = (DWORD)9;
         Opt.AutoRpt = (strncmp(cBuf, "Ye", 2) ? FALSE : TRUE);
			RegQueryValueEx(hKeyFreeCD, "KeepRepeat", 0, NULL, (LPBYTE)&cBuf, (LPDWORD)&dwData);
         dwData = (DWORD)9;
         Opt.KeepRpt = (strncmp(cBuf, "Ye", 2) ? FALSE : TRUE);
			RegQueryValueEx(hKeyFreeCD, "TaskbarTime", 0, NULL, (LPBYTE)&cBuf, (LPDWORD)&dwData);
         dwData = (DWORD)9;
         Opt.TaskTime = (strncmp(cBuf, "Ye", 2) ? FALSE : TRUE);
      }
      else // if the key didn't already exist, just show the window as usual...
      	ShowWindow(hWndMain, SW_SHOW);

  		UpdateWindow(hWndMain); // send a WM_PAINT message to the window
      RegCloseKey(hKeyFreeCD); // close the registry key
   }
}


/////////////////////////// Main function section //////////////////////////////////

// -- add an icon to the taskbar notification area --
// this is called in response to a request to minimize the
// window (see the IDBTN_MIN case under the WM_COMMAND case
// in MainWndProc())
// -- RETURNS whether the icon is succesfully created
BOOL TaskBarAddIcon(LPSTR lpszTip)
{
	BOOL res;
   NOTIFYICONDATA tnid;

   tnid.cbSize = sizeof(NOTIFYICONDATA);
   tnid.hWnd = hWndMain;
   tnid.uID = IDI_TBICON;
   tnid.uFlags = NIF_MESSAGE | NIF_ICON | NIF_TIP;
   tnid.uCallbackMessage = CDWM_NOTIFYICON;

   // display whatever icon is appropriate for the
   // current disc status...
   switch (Status)
   {
   	case playing:
			tnid.hIcon = hIconPlay;
         break;
      case paused:
      	tnid.hIcon = hIconPause;
         break;
      default:
      	tnid.hIcon = hIconX;
   }

	// just in case, for whatever reason, TaskBarAddIcon is
   // called without text for the tooltip, set it to null
	if (lpszTip)
		lstrcpyn(tnid.szTip, lpszTip, sizeof(tnid.szTip));
	else
		tnid.szTip[0] = '\0';

   res = Shell_NotifyIcon(NIM_ADD, &tnid);
   return res;
}

// -- modify current taskbar icon --
// called in TimerProc() to update the current
// time and disc status
// -- RETURNS whether the icon was succesfully changed
BOOL TaskBarModifyIcon(LPSTR lpszTip)
{
   BOOL res;
	NOTIFYICONDATA tnid;

	tnid.cbSize = sizeof(NOTIFYICONDATA);
	tnid.hWnd = hWndMain;
	tnid.uID = IDI_TBICON;
	tnid.uFlags = NIF_MESSAGE | NIF_ICON | NIF_TIP;
   tnid.uCallbackMessage = CDWM_NOTIFYICON;

	// update the icon to reflect the current disc status...
	switch (Status) {
   	case playing:
			tnid.hIcon = hIconPlay;
         break;
      case paused:
      	tnid.hIcon = hIconPause;
         break;
      default:
      	tnid.hIcon = hIconX;
   }

	// make sure the tooltip is set to something...
	if (lpszTip)
		lstrcpyn(tnid.szTip, lpszTip, sizeof(tnid.szTip));
	else
		tnid.szTip[0] = '\0';

	res = Shell_NotifyIcon(NIM_MODIFY, &tnid);
   return res;
}

// -- delete an icon from the taskbar --
// called when the window is restored (un-minimized);
// see the IDBTN_MIN case under the WM_COMMAND case
// in MainWndProc()
// -- RETURNS whether the icon is successfully removed
BOOL TaskBarDeleteIcon()
{
   BOOL res;
	NOTIFYICONDATA tnid;

	tnid.cbSize = sizeof(NOTIFYICONDATA);
	tnid.hWnd = hWndMain;
	tnid.uID = IDI_TBICON;

	res = Shell_NotifyIcon(NIM_DELETE, &tnid);
   return res;
}

// -- prepare the window refresh bitmap --
// this is called when the main window is first created
// (see the WM_CREATE case in MainWndProc());  it creates
// and fills a bitmap in memory to be used to refresh
// the window in response to WM_PAINT messages
void PrimeBmpOps()
{
	HDC hdcMain, hdcMemTemp;
	HBITMAP hBitmap, hbmpOld;
	BITMAP bm;
   RECT rc = {0, 0, WinWidth, WinHeight};

	hdcMain = GetDC(hWndMain);
	hdcMemMain = CreateCompatibleDC(hdcMain);
   hdcMemTemp = CreateCompatibleDC(hdcMemMain);
   hBitmap = CreateBitmap(WinWidth, WinHeight, 1, 16, 0);
   SelectObject(hdcMemMain, hBitmap);
	// fill it up with black to make sure there are no empty areas
   FillRect(hdcMemMain, &rc, (HBRUSH)GetStockObject(BLACK_BRUSH));

	// add the button and display bitmaps...
   for (int i = 0; i <= 9; i++) {
		hBitmap = LoadBitmap(hInst, ref[i].bmpup);
		GetObject(hBitmap, sizeof(BITMAP), &bm);
      hbmpOld = (HBITMAP)SelectObject(hdcMemTemp, hBitmap);
		BitBlt(hdcMemMain, ref[i].x, ref[i].y, bm.bmWidth, bm.bmHeight, hdcMemTemp, 0, 0,  SRCCOPY);
   }

   SelectObject(hdcMemTemp, hbmpOld); // the default object must always be reselected
   DeleteObject(hBitmap);
   DeleteDC(hdcMemTemp);
   ReleaseDC(hWndMain, hdcMain);
}

// -- bitmap output function --
// this is called in OutChars() and during owner-draw
// operations (see the WM_DRAWITEM case in MainWndProc());
// basically defines a generic, multi-purpose bitmap
// drawing procedure
void DrawBmp(int x, int y, LPSTR BmpName)
{
  	HDC hdc, hdcTemp;
	HBITMAP hBitmap, hbmpOld;
	BITMAP bm;

   if (!Hidden) // only do this if the main window is visible (not minimized)
   	hdc = GetDC(hWndMain);
	hdcTemp = CreateCompatibleDC(hdcMemMain);
	hBitmap = LoadBitmap(hInst, BmpName);
	GetObject(hBitmap, sizeof(BITMAP), &bm);
	hbmpOld = (HBITMAP)SelectObject(hdcTemp, hBitmap);
   if (!Hidden) // only if not minimized
		BitBlt(hdc, x, y, bm.bmWidth, bm.bmHeight, hdcTemp, 0, 0,  SRCCOPY);
   // always update the refresh bitmap in memory
   BitBlt(hdcMemMain, x, y, bm.bmWidth, bm.bmHeight, hdcTemp, 0, 0,  SRCCOPY);
   SelectObject(hdcTemp, hbmpOld); // the default object MUST be reselected
   DeleteObject(hBitmap);
   DeleteDC(hdcTemp);
   if (!Hidden) // only if not minimized
   	ReleaseDC(hWndMain, hdc);
}

// -- output track & time characters --
// called in TimerProc(); reads each digit from the
// disc-position string and displays the corresponding
// bitmap (LED digit)
void OutChars(char cPos[12])
{
	char Buffer[50], chr;
	char *pBuf;
   BOOL Zero = FALSE;

   pBuf = cPos;

	chr = *(pBuf + 0); // get the first character of the string
   if (!(chr == '0'))
   { // if it's a character worth displaying (not a 0)...
		wsprintf(Buffer, "Diga%c", chr);
		DrawBmp(ptTrack.x, ptTrack.y, Buffer);
	}
	else
   { // if not, display a blank bitmap to wipe out anything previously put there...
   	Zero = TRUE;
		DrawBmp(ptTrack.x, ptTrack.y, "Digan");
   }

	chr = *(pBuf + 1); // get the next character of the string
   if (Zero && (chr == '0')) // if both of the first two numbers are zeros, display another blank
   	DrawBmp(ptTrack.x + 13, ptTrack.y, "Digan"); // this happens in disc-length-remaining mode
	else
   { // otherwise, display the corresponding bitmap...
		wsprintf(Buffer, "Diga%c", chr);
		DrawBmp(ptTrack.x + 13, ptTrack.y, Buffer);
   }

	chr = *(pBuf + 3); // skip the colon, and get the next character
   if (!(chr == '0'))
   { // if it's not a 0...
   	if ((Disp == remtrk) || (Disp == remdisc)) // if it's in one of the time-remaining modes, display a minus sign
      	DrawBmp(ref[MINUS].x, ref[MINUS].y, "BmpMinus");
      else
      	DrawBmp(ref[MINUS].x, ref[MINUS].y, "Digbn"); // otherwise, display a blank to cover up any previous minus signs
		wsprintf(Buffer, "Digb%c", chr); // display the number
      DrawBmp(ptTime.x, ptTime.y, Buffer);
	}
	else
   { // if it is a 0, move over one bitmap-width to the right and try again...
   	if ((Disp == remtrk) || (Disp == remdisc))
      	DrawBmp(ptTime.x, ptTime.y, "BmpMinus");
      else
			DrawBmp(ptTime.x, ptTime.y, "Digbn"); // cover up any previous minus signs
      DrawBmp(ref[MINUS].x, ref[MINUS].y, "Digbn"); // cover up any previous minus signs
	}

	chr = *(pBuf + 4); // get the next character...
	wsprintf(Buffer, "Digb%c", chr);
	DrawBmp(ptTime.x + 9, ptTime.y, Buffer);

	chr = *(pBuf + 6); // skip the colon again, and get the next character
	wsprintf(Buffer, "Digb%c", chr);
	DrawBmp(ptTime.x + 24, ptTime.y, Buffer);

	chr = *(pBuf + 7); // get the last character
	wsprintf(Buffer, "Digb%c", chr);
	DrawBmp(ptTime.x + 33, ptTime.y, Buffer);
}

// -- get track length info for each track --
// called when a new disc is inserted, or when FreeCD
// first starts (see the WM_CREATE case in MainWndProc());
// fills the array of TrackInfo structures (trk[])
void GetDiscInfo()
{
	char Buffer[50], Buffer2[50], NT[12];

   mciSendString((LPSTR)"status cdaudio current track", Buffer, 25, NULL);
   Track = atoi (Buffer); // set the current track variable
   mciSendString((LPSTR)"status cdaudio number of tracks", NT, 12, NULL);
   NumTracks = atoi (NT); // set the number-of-tracks variable
   mciSendString((LPSTR)"status cdaudio length", disc.str, 6, NULL);
   TimeToInt(disc); // fill the integer members of the TrackInfo structure
   wsprintf(Buffer, "%02d:%s", NumTracks, disc.str);
   OutChars(Buffer); // display the number of tracks and total disc length

   // the following two lines will be used in the future to associate
   // info for each disc with a unique disc identifier in the soon-to-be
   // implemented disc-info database:
   //
   //-> mciSendString((LPSTR)"info cdaudio identity", Buffer, 20, NULL);
   //-> MessageBox(hWndMain, Buffer, "Disc ID#", MB_OK);

   // get the length of each track and fill the integer member of
   // the respective TrackInfo structures
   for (int i = 0; i < NumTracks; i++){
   	wsprintf(Buffer2, "status cdaudio length track %d", (i + 1));
   	mciSendString((LPSTR)Buffer2, trk[i].str, 6, NULL);
      TimeToInt(trk[i]);
   }
}

// -- fill the integer members of a TrackInfo structure --
// this transfers a TrackInfo structure's string member
// to minute & second integers
void TimeToInt(TrackInfo &tr)
{
   tr.m = atoi(tr.str);
   tr.s = atoi(tr.str + 3);
}

// -- reverse track skip --
// called when then corresponding button is pressed,
// or when a track needs to be repeated
// NOTE: a default value is specified so this can be
// called with no parameters...
BOOL RevSkip(BOOL RepeatTrack = FALSE)
{
   char Buffer[50];

   // if this WASN'T called in response to a track repeat
   // (that is, it was called because the reverse-skip
   // button was pressed)...
	if (!RepeatTrack)
   {
		mciSendString((LPSTR)"status cdaudio current track", Buffer, 25, NULL);
		Track = atoi (Buffer);
   	TimeToInt(cur);
      // if the current position is within the first 5 seconds of the track,
      // skip to the previous track; otherwise, skip to the beginning of the
      // current track...
   	if ((cur.m == 0) && (cur.s < 5))
      {
			if (Track > 1)
				Track--;
			else // if reverse-skipping from track #1, skips to the last track on the disc
   			Track = NumTracks;
         if (!Opt.KeepRpt)
         { // if KeepRepeat isn't enabled and Repeat is on, turn off Repeat
         	Repeat = FALSE;
            UpdatePlayBtn(); // update the repeat button to reflect the new status
         }
         else // otherwise, change the track to repeat to the newly selected track
         	RptTrack = Track;
   	}
   }
	// if it WAS called to repeat the track and the disc accidently kept
   // playing into the next track, change the current track to the track
   // that was originally meant to be repeated
   else if (Track != RptTrack)
   	Track = RptTrack;
	wsprintf(Buffer, "play cdaudio from %d", Track);
	if(!mciSendString((LPSTR)Buffer, NULL, NULL, NULL))
   	return TRUE;
   return FALSE;
}

// -- play the disc --
// called when the user presses the play/pause button;
// if the disc is currently paused or stopped, this plays it
// RETURNS whether the MCI play command succeeds
BOOL PlayDisc()
{
	char Buffer[50];

	// use the most-recent disc position, which will
	// be the last-saved position if the disc is
   // "paused", or 01:00:00:00 if the disc is stopped
	strcpy(Buffer, "play cdaudio from ");
	strncat(Buffer, Pos, 11);
	if(!mciSendString((LPSTR)Buffer, NULL, NULL, NULL))
   	return TRUE;
   return FALSE;
}

// -- pause the disc --
// called when the play/pause button is pressed and the
// disc is currently playing; saves the position where
// the disc is paused, since in the Windows MCI interface,
// there is no real pause command (that is, even the so-
// called "pause" command does the same thing as the "stop"
// command)
void PauseDisc()
{
	char Buffer[50];

	mciSendString((LPSTR)"stop cdaudio", NULL, NULL, NULL);
	mciSendString((LPSTR)"status cdaudio position", Buffer, 25, NULL);
	strncpy(Pos, Buffer, 11);
}

// -- forward track-skip --
// skips to the next track
BOOL FFSkip()
{
   char Buffer[50];

	mciSendString((LPSTR)"status cdaudio current track", Buffer, 25, NULL);
	Track = atoi (Buffer);
	if (Track < NumTracks)
		Track++;
	else // if forward-skipping from the last track on the disc, skips to track #1
		Track = 1;
   if (!Opt.KeepRpt)
   { // if KeepRepeat is disabled, turn off Repeat
   	Repeat = FALSE;
      UpdatePlayBtn(); // update the repeat button to reflect the new status
   }
   else // if KeepRepeat is enabled, change the track-to-repeat to the newly selected track
   	RptTrack = Track;
	wsprintf(Buffer, "play cdaudio from %d", Track);
	if(!mciSendString((LPSTR)Buffer, NULL, NULL, NULL))
   	return TRUE;
   return FALSE;
}

// -- process messages for taskbar icon --
// this is called when MainWndProc() receives the
// application-defined CDWM_NOTIFYICON message to
// indicate that the icon has received some input
void On_CDWM_NOTIFYICON(LPARAM lParam)
{
	UINT uMouseMsg;
   uMouseMsg = (UINT) lParam;

	if (uMouseMsg == WM_RBUTTONDOWN) // if the right mouse button was clicked...
   	DisplayTaskBarMenu();
   else if (uMouseMsg == WM_LBUTTONDOWN) // if the left mouse button was clicked...
   {
		// perform whatever action corresponds to the current icon displayed...
   	switch (Status)
      {
        	case playing:
         	// if the playing icon is displayed, pause the disc
				PauseDisc();
				Status = paused;
           	break;

         case open:
         	// if the 'X' icon is displayed, simulate pressing the
            // eject button to close the drive
         	PostMessage(hWndMain, WM_COMMAND, IDBTN_EJ, NULL);
         	break;

         default:
         	// otherwise, play the disc
				PlayDisc();
				Status = playing;
            break;
		}
   }
	return;
}

// -- process WM_CONTEXTMENU message --
// this is called when the right mouse button is
// clicked in the main window to display the menu
// -- RETURNS whether the mouse was clicked in
// the window's client area
BOOL WINAPI OnContextMenu(int x, int y)
{
    RECT rc;
    POINT pt = { x, y };

    GetClientRect(hWndMain, &rc);
    ScreenToClient(hWndMain, &pt);
    if (PtInRect(&rc, pt))
    {
        ClientToScreen(hWndMain, &pt);
        DisplayContextMenu(pt);
        return TRUE;
    }
    return FALSE;
}

// -- display main window menu --
// called in response to OnContextMenu();
// displays the popup menu
void APIENTRY DisplayContextMenu(POINT pt)
{
   TPMPARAMS tpm;
   RECT rc;

	GetWindowRect(hWndMain, &rc);
   tpm.cbSize = sizeof(TPMPARAMS);
   tpm.rcExclude = rc; // make sure the menu doesn't overlap any part of the window
	TrackPopupMenuEx(hMenuMain, TPM_CENTERALIGN | TPM_RIGHTBUTTON | TPM_VERTICAL,
            pt.x, pt.y, hWndMain, (LPTPMPARAMS)&tpm);
}

// -- display taskbar menu --
// called in response to On_CDWM_NOTIFYICON();
// displays the popup menu for the taskbar icon
void APIENTRY DisplayTaskBarMenu()
{
   int x, y;
   TPMPARAMS tpm;
	APPBARDATA abd;

   abd.cbSize = sizeof(APPBARDATA);
   abd.hWnd = hWndMain;
   SHAppBarMessage(ABM_GETTASKBARPOS, &abd);
   x = abd.rc.right;
   y = abd.rc.bottom;
   tpm.cbSize = sizeof(TPMPARAMS);
   tpm.rcExclude = abd.rc; // make sure the menu doesn't overlap the taskbar
   SetForegroundWindow(hWndMain); // this is NECESSARY to override a bug in Windows
	TrackPopupMenuEx(hMenuMain, TPM_LEFTALIGN | TPM_RIGHTBUTTON | TPM_VERTICAL,
            x, y, hWndMain, (LPTPMPARAMS)&tpm);
   PostMessage(hWndMain, WM_USER, NULL, NULL); // this is NECESSARY to override a bug in Windows
}

// -- prepare to exit --
// called before FreeCD closes to allow settings
// to be saved and resources to be freed
void KillApp()
{
   HKEY hKeyFreeCD;
   DWORD dwResult;
	char cWinX[8] = "",
   	  cWinY[8] = "",
        cHidden[8] = "",
   	  cTBIcon[8] = "",
        cOnTop[8] = "",
        cScrnPos[8] = "",
   	  cAutoPlay[8] = "",
        cAutoRpt[8] = "",
        cKeepRpt[8] = "",
        cTaskTime[8] = "";

   itoa(WinX, cWinX, 10);
   itoa(WinY, cWinY, 10);
   strcpy(cHidden, Hidden ? "Yes" : "No");
   strcpy(cTBIcon, Opt.TBIcon ? "Yes" : "No");
   strcpy(cOnTop, TopMost ? "Yes" : "No");
   strcpy(cScrnPos, Opt.ScrnPos ? "Yes" : "No");
   strcpy(cAutoPlay, Opt.AutoPlay ? "Yes" : "No");
   strcpy(cAutoRpt, Opt.AutoRpt ? "Yes" : "No");
   strcpy(cKeepRpt, Opt.KeepRpt ? "Yes" : "No");
   strcpy(cTaskTime, Opt.TaskTime ? "Yes" : "No");

   // open the settings key in the registry
	RegCreateKeyEx(HKEY_CURRENT_USER, "Software\\Ion\\FreeCD", 0,
   					NULL, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL,
                  &hKeyFreeCD, &dwResult);

	// write the new values
	RegSetValueEx(hKeyFreeCD, "AlwaysMinimized", 0, REG_SZ, (LPBYTE)cTBIcon, strlen(cTBIcon) + 1);
   RegSetValueEx(hKeyFreeCD, "AlwaysOnTop", 0, REG_SZ, (LPBYTE)cOnTop, strlen(cOnTop) + 1);
	RegSetValueEx(hKeyFreeCD, "AutoPlay", 0, REG_SZ, (LPBYTE)cAutoPlay, strlen(cAutoPlay) + 1);
	RegSetValueEx(hKeyFreeCD, "AutoRepeat", 0, REG_SZ, (LPBYTE)cAutoRpt, strlen(cAutoRpt) + 1);
	RegSetValueEx(hKeyFreeCD, "KeepRepeat", 0, REG_SZ, (LPBYTE)cKeepRpt, strlen(cKeepRpt) + 1);
	RegSetValueEx(hKeyFreeCD, "Minimized", 0, REG_SZ, (LPBYTE)cHidden, strlen(cHidden) + 1);
	RegSetValueEx(hKeyFreeCD, "TaskbarTime", 0, REG_SZ, (LPBYTE)cTaskTime, strlen(cTaskTime) + 1);
	RegSetValueEx(hKeyFreeCD, "UseLastWindowPos", 0, REG_SZ, (LPBYTE)cScrnPos, strlen(cScrnPos) + 1);
	RegSetValueEx(hKeyFreeCD, "WindowPosX", 0, REG_SZ, (LPBYTE)cWinX, strlen(cWinX) + 1);
	RegSetValueEx(hKeyFreeCD, "WindowPosY", 0, REG_SZ, (LPBYTE)cWinY, strlen(cWinY) + 1);

   RegCloseKey(hKeyFreeCD);

   if(wTimerID) // if the timer exists, kill it
   {
   	timeKillEvent(wTimerID);
      timeEndPeriod(wTimerRes);
	}
   TaskBarDeleteIcon(); // if minimized, remove the taskbar icon
	mciSendString((LPSTR)"close cdaudio", NULL, NULL, NULL); // close the cdaudio device
   DestroyMenu(hMenuSys); // free the system menu resources
   DestroyMenu(hMenuMain); // free the main menu resouces
   DestroyIcon(hIconPlay);		// free the icon resources...
   DestroyIcon(hIconPause);	//...
   DestroyIcon(hIconX);			//...
   DeleteDC(hdcMemMain); // free the window-refresh bitmap resources
   DestroyWindow(hWndMain); // destroy the main window
}

// -- update the button status --
// this is called whenever the disc status (play/pause/stop)
// changes or when Repeat is turned on & off;  redraws
// the buttons to reflect their new status
void UpdatePlayBtn()
{
	switch (Status)
   {
      case paused:
      	DrawBmp(ref[PLAY].x, ref[PLAY].y, "BtnPauseUp");
         break;
		case playing:
      	DrawBmp(ref[PLAY].x, ref[PLAY].y, "BtnPlayUp");
         break;
		default:
      	DrawBmp(ref[PLAY].x, ref[PLAY].y, "BtnStopUp");
         break;
   }
   if (Repeat)
   	DrawBmp(ref[RPT].x, ref[RPT].y, "BtnRptAct");
   else
   	DrawBmp(ref[RPT].x, ref[RPT].y, "BtnRptUp");
}

// -- draw owner-drawn menu & dialog bitmap --
// this draws the "FreeCD" logo banner along the left-hand
// side of the menus and dialog boxes; it's similar to
// DrawBmp() but has a few things changed/added...
void OwnerDrawBmp(LPDRAWITEMSTRUCT lpdis)
{
	HWND hWndTemp;
   HDC hdcTemp;
   HBITMAP hBitmap, hbmpOld;
   BITMAP bm;
   RECT rc;
   int x, y, Height, cury, bmh;

	// set the top-left corner of the bitmap to that
   // returned in the DRAWITEMSTRUCT
   x = lpdis->rcItem.left;
   y = lpdis->rcItem.top;

   hWndTemp = WindowFromDC(lpdis->hDC); // use the DC from the DRAWITEMSTRUCT
   GetClientRect(hWndTemp, &rc); // get the client rectangle of the DC
   // add 1, since a RECT structure doesn't include
   // the right and bottom edges of the rectangle
   Height = rc.bottom - rc.top + 1;

   hdcTemp = CreateCompatibleDC(lpdis->hDC);
   hBitmap = LoadBitmap(hInst, "BmpLogo");
   GetObject(hBitmap, sizeof(BITMAP), &bm);
   hbmpOld = (HBITMAP)SelectObject(hdcTemp, hBitmap);
   BitBlt(lpdis->hDC, x, y, bm.bmWidth, bm.bmHeight, hdcTemp, 0, 0, SRCCOPY);
   SelectObject(hdcTemp, hbmpOld); // the default object MUST be reselected
   DeleteObject(hBitmap);

	// if the height of the destination rectangle is greater than the
   // height of the logo bitmap (BmpLogo), keep adding the filler
   // bitmap (BmpLogo2) onto the bottom until the destination
   // rectangle is filled
   if (Height > 94)
   {
   	cury = 94;
      hBitmap = LoadBitmap(hInst, "BmpLogo2");
      GetObject(hBitmap, sizeof(BITMAP), &bm);
      hbmpOld = (HBITMAP)SelectObject(hdcTemp, hBitmap);
      do {
      	bmh = ((Height - cury) > 1) ? bm.bmHeight : (Height - cury);
         BitBlt(lpdis->hDC, x, y + cury, bm.bmWidth, bmh, hdcTemp, 0, 0, SRCCOPY);
         cury += 1;
      } while ((Height - cury) > 1);
      SelectObject(hdcTemp, hbmpOld); // reselect the default object
      DeleteObject(hBitmap);
   }
   DeleteDC(hdcTemp);
}

// -- draw the owner-drawn buttons --
// this is called in response to the WM_DRAWITEM case
// in MainWndProc(); it just helps cut down on the
// repetitive code
void DrawBtn(LPDRAWITEMSTRUCT lpdis, int btn)
{
	if (lpdis->itemState == ODS_FOCUS)
		DrawBmp(ref[btn].x, ref[btn].y, ref[btn].bmpup);
	else
		DrawBmp(ref[btn].x, ref[btn].y, ref[btn].bmpdn);
}


///////////////////////// Dialog Callback Functions ///////////////////////////////

// -- handle messages for the Help dialog --
#pragma argsused
BOOL CALLBACK HelpDlgProc(HWND hWndDlg, UINT message,
								  WPARAM wParam, LPARAM lParam)
{
	RECT rc, rcDlg;
	int DlgX, DlgY;

	switch (message) {
		case WM_INITDIALOG:
      	// center the dialog on the screen
         GetWindowRect(GetDesktopWindow(), &rc);
         GetWindowRect(hWndDlg, &rcDlg);
         rcDlg.right = rcDlg.right - rcDlg.left;
         rcDlg.bottom = rcDlg.bottom - rcDlg.top;
         DlgX = INT((rc.right - rcDlg.right) / 2);
			DlgY = INT((rc.bottom - rcDlg.bottom) / 2);

         SetWindowPos(hWndDlg, NULL, DlgX, DlgY, 0, 0, SWP_NOSIZE | SWP_NOZORDER);
      	return TRUE;

      case WM_DRAWITEM:
      	// draw the logo banner along the left side of the dialog
      	LPDRAWITEMSTRUCT lpdis;

         lpdis = (DRAWITEMSTRUCT FAR*) lParam;
         if (lpdis->CtlID == IDS_BMP)
            OwnerDrawBmp(lpdis);
         return TRUE;

		case WM_COMMAND:
      	// if either the "Thanks" button or the little 'X' in
         // the top-right corner is clicked, close the dialog
      	switch (wParam)
         {
            case IDOK:
            case IDCANCEL:
            	DestroyWindow(hWndDlg);
               // reset the handle to NULL so that MainWndProc() will
               // know whether the Help dialog already exists
               hWndHlp = NULL;
               return TRUE;
      	}
	}
	return FALSE;
}

// -- handle messages for the Options dialog --
#pragma argsused
BOOL CALLBACK OptDlgProc(HWND hWndDlg, UINT message,
								  WPARAM wParam, LPARAM lParam)
{
	RECT rc, rcDlg;
	int DlgX, DlgY;
	static OptionsStruct Temp;

	switch (message)
   {
		case WM_INITDIALOG:
      	// center the dialog on the screen
         GetWindowRect(GetDesktopWindow(), &rc);
         GetWindowRect(hWndDlg, &rcDlg);
         rcDlg.right = rcDlg.right - rcDlg.left;
         rcDlg.bottom = rcDlg.bottom - rcDlg.top;
         DlgX = INT((rc.right - rcDlg.right) / 2);
			DlgY = INT((rc.bottom - rcDlg.bottom) / 2);

         SetWindowPos(hWndDlg, NULL, DlgX, DlgY, 0, 0, SWP_NOSIZE | SWP_NOZORDER);

			// copy the options into a temporary structure, and
         // set the state of the check controls to reflect the
         // current settings
         Temp = Opt;
         if (Temp.TaskTime)
            CheckDlgButton(hWndDlg, IDCHK_TASKTIME, BST_CHECKED);
         if (Temp.TBIcon)
            CheckDlgButton(hWndDlg, IDCHK_TBICON, BST_CHECKED);
         if (Temp.ScrnPos)
            CheckDlgButton(hWndDlg, IDCHK_SCRNPOS, BST_CHECKED);
         if (Temp.AutoPlay)
            CheckDlgButton(hWndDlg, IDCHK_AUTOPLAY, BST_CHECKED);
         if (Temp.AutoRpt)
            CheckDlgButton(hWndDlg, IDCHK_AUTORPT, BST_CHECKED);
         if (Temp.KeepRpt)
            CheckDlgButton(hWndDlg, IDCHK_KEEPRPT, BST_CHECKED);

      	return TRUE;

      case WM_DRAWITEM:
      	// draw the banner bitmap along the left-hand side of the dialog
      	LPDRAWITEMSTRUCT lpdis;

         lpdis = (DRAWITEMSTRUCT FAR*) lParam;
         if (lpdis->CtlID == IDS_BMP)
            OwnerDrawBmp(lpdis);
         return TRUE;

		// process control messages...
		case WM_COMMAND:
      	switch (wParam)
         {
            case IDOK: // OK button was pressed
            	Opt = Temp; // save the new settings
             	DestroyWindow(hWndDlg);
               hWndOptions = NULL; // reset the handle to NULL
               return TRUE;

            case IDCANCEL: // Cancel button was pressed
            	// don't save the new settings;
					// if the time was shown in the taskbar button and isn't now
               // (or vice-versa), reset it to its original state
               if (Opt.TaskTime != PrevTime)
               {
	              	Opt.TaskTime = PrevTime;
                  if (Opt.TaskTime)
		              	SetWindowText(hWndMain, SPos);
                  else
                  	SetWindowText(hWndMain, "FreeCD");
               }
	           	DestroyWindow(hWndDlg);
               hWndOptions = NULL; // reset the handle
               return TRUE;

            // the rest respond to check controls...
         	case IDCHK_TASKTIME:
            	if (Temp.TaskTime)
               {
               	// show change immediately so the user
                  // can see the difference
               	Temp.TaskTime = FALSE;
                  Opt.TaskTime = FALSE;
                  SetWindowText(hWndMain, "FreeCD");
               }
               else
               {
               	// show the change immediately so the
                  // user can see the difference
               	Temp.TaskTime = TRUE;
                  Opt.TaskTime = TRUE;
                  SetWindowText(hWndMain, SPos);
               }
               break;

         	case IDCHK_TBICON:
            	if (Temp.TBIcon)
               	Temp.TBIcon = FALSE;
               else
               	Temp.TBIcon = TRUE;
               break;

         	case IDCHK_SCRNPOS:
            	if (Temp.ScrnPos)
               	Temp.ScrnPos = FALSE;
               else
               	Temp.ScrnPos = TRUE;
               break;

         	case IDCHK_AUTOPLAY:
            	if (Temp.AutoPlay)
               	Temp.AutoPlay = FALSE;
               else
               	Temp.AutoPlay = TRUE;
               break;

         	case IDCHK_AUTORPT:
            	if (Temp.AutoRpt)
               	Temp.AutoRpt = FALSE;
               else
               	Temp.AutoRpt = TRUE;
               break;

         	case IDCHK_KEEPRPT:
            	if (Temp.KeepRpt)
               	Temp.KeepRpt = FALSE;
               else
               	Temp.KeepRpt = TRUE;
               break;
      	}
	}
	return FALSE;
}


//////////////////////////// Timer Callback Procedure //////////////////////////////
//
// -- perform routine status check and display update --
// this is called every <TimeOut> milliseconds by the
// call to timeSetEvent() in the WM_CREATE case in
// MainWndProc()
#pragma argsused
void FAR PASCAL TimerProc(UINT wTimerID, UINT msg, DWORD dwUser, DWORD dw1, DWORD dw2)
{
	char Buffer[50];
	char CPos[9], CPos2[9] = "01:00:00";
   char *pchr;

	// if a wait period is in effect, exit the procedure
   // and wait another <TimeOut> milliseconds
	if (Wait)
   {
   	Wait--;
      if (Wait >= 1)
      	return;
   }

   // check if there's a disc in the drive
   mciSendString((LPSTR)"status cdaudio media present", Buffer, 5, NULL);
   if (strncmp(Buffer, "true", 4))
   { // if not...
   	if (DiscPres == TRUE)
      { // if it was just removed...
      	Status = open;
	      UpdatePlayBtn(); // update the play button to reflect the new state
         DiscPres = FALSE;
         // change which bitmap to use as the current default
         strcpy(ref[DISP].bmpup, "BmpNoDisc");
         DrawBmp(ref[DISP].x, ref[DISP].y, ref[DISP].bmpup);
         if (Hidden) // if minimized...
	         TaskBarModifyIcon("no disc  [FreeCD]");
         if (Opt.TaskTime) // if the time should be displayed in the taskbar button...
	        	SetWindowText(hWndMain, "no disc  [FreeCD]");
         Track = 1; // reset the current track
      }
      return;
   }
   else if (DiscPres == FALSE)
   { // or if it was just inserted...
   	DiscPres = TRUE;
      strcpy(CPos, "01:00:00");
      strcpy(Pos, "01:00:00");
      if (Opt.AutoPlay)
      {
	      PlayDisc();
   	   Status = playing;
	      UpdatePlayBtn();
		}
      strcpy(ref[DISP].bmpup, "BmpClrTime");
      DrawBmp(ref[DISP].x, ref[DISP].y, ref[DISP].bmpup); // clear the time display
      GetDiscInfo();
      // make sure the disc length is displayed for
      // at least a couple seconds...
      Wait = 2;
      return;
   }

   // if the disc has reached the end, repeat it
   mciSendString((LPSTR)"status cdaudio mode", Buffer, 8, NULL);
 	if ( (!strncmp(Buffer, "stopped", 7)) && (Status == playing) )
	{
   	if (Opt.AutoRpt)
   	{
   		strcpy(Pos, "01:00:00");
	      PlayDisc();
	   }
      else
      {
      	Status = open;
         UpdatePlayBtn();
         Disp = norm;
		   wsprintf(Buffer, "%02d:%s", NumTracks, disc.str);
		  	OutChars(Buffer);
      }
   }

   // get the current disc position...
   if (Status != open)
   {
 		mciSendString((LPSTR)"status cdaudio position", CPos, 9, NULL);
     	Track = atoi(CPos);
      pchr = strchr(CPos, ':'); // find the first colon
      pchr++; // move the pointer to the next character
      strcpy(cur.str, pchr); // copy the remaining characters
      TimeToInt(cur); // copy the string info into the integer members of the structure
      if (Repeat) // if Repeat is on...
      {
			// if the current position is within the last second
         // of the track, or the current track is NOT the same
         // as the track which was originally intended to be
         // repeated, skip to the beginning of the track to be
         // repeated...
      	if (   ((cur.m == trk[Track-1].m) && (cur.s >= trk[Track-1].s-1))   ||   (Track != RptTrack)   )
         {
      		RevSkip(TRUE);
            return;
         }
      }
		if (Disp == remdisc)
      {
      	// if in disc-length-remaining mode, get the current
      	// position in minute:second:frame format, rather
         // than track:minute:second:frame format...
      	mciSendString((LPSTR)"set cdaudio time format msf", NULL, NULL, NULL);
         mciSendString((LPSTR)"status cdaudio position", CPos2, 9, NULL);
         mciSendString((LPSTR)"set cdaudio time format tmsf", NULL, NULL, NULL);
      }

      // what's the current display mode?
      switch (Disp)
      {
      	case norm: // if normal display mode...
     	     	OutChars(CPos); // output the time display string as-is...
   	   	wsprintf(SPos, "%d  %d:%02d  [FreeCD]", Track, cur.m, cur.s);
            if (Hidden) // if minimized, update the taskbar icon tooltip
               TaskBarModifyIcon(SPos);
		      else if (Opt.TaskTime) // if TaskTime enabled, update the taskbar button text
	      	   SetWindowText(hWndMain, SPos);
            break;

         case remtrk: // if track-length-remaining mode...
	      	TimeToInt(cur); // copy the string info to the integer members of the structure
            // calculate the time remaining for the track...
   	      rem.m = trk[Track-1].m - cur.m;
      	   rem.s = trk[Track-1].s - cur.s;
         	if (rem.s < 0)
            {
	         	rem.m--;
   	         rem.s += 60;
      	   }
            wsprintf(rem.str, "%02d:%02d", rem.m, rem.s);
  	      	wsprintf(Buffer, "%02d:", Track);
	         strncat(Buffer, rem.str, 5);
		     	OutChars(Buffer);
  	      	wsprintf(SPos, "%d  -%d:%02d  [FreeCD]", Track, rem.m, rem.s);
		      if (Hidden)
   		      TaskBarModifyIcon(SPos);
		      else if (Opt.TaskTime)
   		      SetWindowText(hWndMain, SPos);
            break;

         case remdisc: // if disc-length-remaining mode...
            strcpy(cur2.str, CPos2); // copy the current position (in msf format)...
            TimeToInt(cur2); // and copy the string to the integers
            // calculate the time remaining...
            rem.m = disc.m - cur2.m;
            rem.s = disc.s - cur2.s;
            if (rem.s < 0)
            {
            	rem.m--;
               rem.s += 60;
            }
            wsprintf(rem.str, "%02d:%02d", rem.m, rem.s);
            wsprintf(Buffer, "00:%s", rem.str);
		     	OutChars(Buffer);
	      	wsprintf(SPos, "-%d:%02d  [FreeCD]", rem.m, rem.s);
		      if (Hidden)
   			   TaskBarModifyIcon(SPos);
		      else if (Opt.TaskTime)
   			   SetWindowText(hWndMain, SPos);
         	break;

         case total: // if total-disc-length mode...
			   wsprintf(Buffer, "%02d:%s", NumTracks, disc.str);
		     	OutChars(Buffer);
	      	wsprintf(SPos, "%d  %d:%02d  [FreeCD]", NumTracks, disc.m, disc.s);
		      if (Hidden)
   		      TaskBarModifyIcon(SPos);
		      else if (Opt.TaskTime)
   		      SetWindowText(hWndMain, SPos);
         	break;
      }
   }
}


//////////////////////////// Main Callback Procedure /////////////////////////////
//
// main app callback message-handling procedure
LRESULT CALLBACK MainWndProc(HWND hWndMain, UINT message,
										 WPARAM wParam, LPARAM lParam)
{
	char Buffer[50];
	static PAINTSTRUCT ps;

	switch(message)
	{
		case WM_MEASUREITEM: // the owner-drawn buttons are being created...
   		LPMEASUREITEMSTRUCT lpmisCtl;
         int nIDCtl;
         HBITMAP hBitmap;
         BITMAP bm;

         lpmisCtl = (MEASUREITEMSTRUCT FAR*) lParam;
         nIDCtl = (int) wParam;

         if (!nIDCtl) // if there's no ID, it signifies a menu item...
         {
         	hBitmap = LoadBitmap(hInst, "BmpLogo");
            GetObject(hBitmap, sizeof(BITMAP), &bm);
            // copy the bitmaps dimensions into the
            // MEASUREITEMSTRUCT for internal processing;
            // the value of 16 used below can be changed
            // to suit personal preferences (change it and
            // see what happens)
            lpmisCtl->itemWidth = bm.bmWidth - 16;
            lpmisCtl->itemHeight = bm.bmHeight;
            DeleteObject(hBitmap);
         }
         else // otherwise, it's a button...
         {
         	lpmisCtl->CtlType = ODT_BUTTON;
            lpmisCtl->CtlID = nIDCtl;
         }
			return TRUE;


      case WM_DRAWITEM: // the owner-drawn buttons need to be updated...
      	LPDRAWITEMSTRUCT lpdis;

         lpdis = (DRAWITEMSTRUCT FAR*) lParam;
         if (lpdis->CtlID){ // if it has an ID, it's a button...
         	switch (lpdis->itemAction) // what is the requested drawing action?
            {
            	case ODA_DRAWENTIRE: // draw the entire button (called when they're first created)...
						switch (lpdis->CtlID) // which button?
                  {
                     case IDBTN_PLAY: // update the play button based on the current state...
                     	if (Status == playing)
                        	DrawBmp(ref[PLAY].x, ref[PLAY].y, "BtnPlayUp");
                        else if (Status == paused)
                        	DrawBmp(ref[PLAY].x, ref[PLAY].y, "BtnPauseUp");
                        else
                        	DrawBmp(ref[PLAY].x, ref[PLAY].y, ref[PLAY].bmpup);
                        break;

                     case IDBTN_RPT: // update the repeat button based on the current state...
                     	if (Repeat)
                        	DrawBmp(ref[RPT].x, ref[RPT].y, "BtnRptAct");
                        else
                        	DrawBmp(ref[RPT].x, ref[RPT].y, ref[RPT].bmpup);
                        break;

							// all the rest can use the generic function call...
							case IDBTN_REW:
                     case IDBTN_FF:
                     case IDBTN_VOL:
                     case IDBTN_MIN:
                     case IDBTN_EJ:
                     case IDBTN_EXIT:
                     case IDBTN_HELP:
                        int i;
                        i = lpdis->CtlID - 1;
                     	DrawBmp(ref[i].x, ref[i].y, ref[i].bmpup);
                        break;
                  }
                  break;

               case ODA_SELECT: // the button's state has changed...
            		switch (lpdis->CtlID) // which button?
						{
                     case IDBTN_PLAY:
                     	if (lpdis->itemState == ODS_FOCUS) // the button is currently up...
                        {
                        	if (Status == playing)
                           	DrawBmp(ref[PLAY].x, ref[PLAY].y, "BtnPlayUp");
                           else if (Status == paused)
                           	DrawBmp(ref[PLAY].x, ref[PLAY].y, "BtnPauseUp");
                           else
                           	DrawBmp(ref[PLAY].x, ref[PLAY].y, "BtnStopUp");
                        }
                        else // the button is currently pressed down...
                        	DrawBmp(ref[PLAY].x, ref[PLAY].y, "BtnPlayDn");
                        break;

                     case IDBTN_RPT:
                     	if (lpdis->itemState == ODS_FOCUS) // the button is currently up...
                        {
                        	if (Repeat)
                           	DrawBmp(ref[RPT].x, ref[RPT].y, "BtnRptAct");
                           else
                           	DrawBmp(ref[RPT].x, ref[RPT].y, "BtnRptUp");
                        }
                        else // the button is currently pressed down...
                        	DrawBmp(ref[RPT].x, ref[RPT].y, "BtnRptDn");
                        break;

							// all the rest use the generic function...
                  	case IDBTN_REW:
                     case IDBTN_FF:
                     case IDBTN_VOL:
                     case IDBTN_MIN:
                     case IDBTN_EJ:
                     case IDBTN_EXIT:
                     case IDBTN_HELP:
                  		DrawBtn(lpdis, lpdis->CtlID - 1);
                        break;

                     default:
                     	break;
                  }
                  break;
            }
         }
         else // if it doesn't have an ID, it's a menu item, and therefore is the banner bitmap...
            OwnerDrawBmp(lpdis);
         return TRUE;


      case WM_COMMAND: // a button has been pushed...
      case WM_SYSCOMMAND: // or a system-menu command has been clicked...
      	switch(wParam) // which button (or system command)?
         {
         	case IDBTN_REW: // reverse skip...
            	if(RevSkip())
               	Status = playing;
               UpdatePlayBtn();
               break;

            case IDBTN_PLAY: // play/pause...
            	// make sure the disc plays before updating the status
            	if ((Status != playing) && PlayDisc())
                 	Status = playing;
               else if (Status == playing)
               {
               	PauseDisc();
                  Status = paused;
               }
               UpdatePlayBtn();
               break;

            case IDBTN_FF: // forward skip
            	if(FFSkip())
						Status = playing;
					UpdatePlayBtn();
               break;

            case IDBTN_HELP: // help...
            	if (!IsWindow(hWndHlp)) // if the dialog doesn't already exist, create it...
               {
               	hWndHlp = CreateDialog(hInst,
                  	MAKEINTRESOURCE(DLGHLPABOUT),
                     hWndMain, (DLGPROC) HelpDlgProc);
                  ShowWindow(hWndHlp, SW_SHOW);
               }
               else // if it does exist, move it to the foreground...
               	SetForegroundWindow(hWndHlp);
               // if not minimized, repaint the window, just in case
               // it's partially hidden by the new dialog...
               if(!Hidden)
               	RedrawWindow(hWndMain, NULL, NULL, RDW_INTERNALPAINT);
               break;

            case IDBTN_OPT: // options...
            	if (!IsWindow(hWndOptions)) // if the dialog doesn't already exist, create it...
               {
               	PrevTime = Opt.TaskTime; // save the original state of TaskTime
               	hWndOptions = CreateDialog(hInst,
                  	MAKEINTRESOURCE(DLGOPTIONS),
                     hWndMain, (DLGPROC) OptDlgProc);
                  ShowWindow(hWndOptions, SW_SHOW);
               }
            	else // if it does exist, move it to the foreground...
               	SetForegroundWindow(hWndOptions);
               // if not minimized, repaint the window, just in case
               // it's partially hidden by the new dialog...
               if(!Hidden)
               	RedrawWindow(hWndMain, NULL, NULL, RDW_INTERNALPAINT);
               break;

            case IDBTN_RPT: // track repeat...
            	if (!Hidden)
               {
               	if (Repeat)
                  	Repeat = FALSE;
                  else
                  {
                  	Repeat = TRUE;
							mciSendString((LPSTR)"status cdaudio position", Buffer, 9, NULL);
                     RptTrack = atoi(Buffer); // save the track to repeat
                  }
               }
               else // if minimized, check the corresponding menu item to reflect the repeat state
               {
               	if (Repeat)
                  {
                  	Repeat = FALSE;
                     CheckMenuItem(hMenuMain, IDBTN_RPT, MF_BYCOMMAND | MF_UNCHECKED);
                  }
                  else
                  {
                  	Repeat = TRUE;
                     CheckMenuItem(hMenuMain, IDBTN_RPT, MF_BYCOMMAND | MF_CHECKED);
                  }
               }
               UpdatePlayBtn();
               break;

            case IDBTN_EXIT: // exit...
            	KillApp();
               break;

            case IDBTN_VOL: // volume...
            	// if the volume window is already open, move it to the
               // foreground; otherwise, open it
            	if (!SetForegroundWindow(FindWindow(NULL,(LPCTSTR)"Volume Control")))
               	WinExec("sndvol32.exe", SW_SHOWNORMAL);
               break;

            case IDBTN_MIN: // minimize to taskbar...
            	if (!Hidden) // if not currently minimized...
               {
               	Hidden = TRUE;
                  ShowWindow(hWndMain, SW_HIDE); // hide the main window
                  // change the menu for use with the taskbar icon...
                  DeleteMenu(hMenuMain, IDBTN_ONTOP, MF_BYCOMMAND);
                  ModifyMenu(hMenuMain, IDBTN_MIN, MF_BYCOMMAND | MF_STRING, IDBTN_MIN, "Restore window");
                  InsertMenu(hMenuMain, IDBTN_EXIT, MF_BYCOMMAND | MF_STRING, IDBTN_EJ, "Eject");
						InsertMenu(hMenuMain, IDBTN_EXIT, MF_BYCOMMAND | MF_STRING, IDBTN_RPT, "Repeat track");
                  if (Repeat)
                  	CheckMenuItem(hMenuMain, IDBTN_RPT, MF_BYCOMMAND | MF_CHECKED);
                  InsertMenu(hMenuMain, IDBTN_EXIT, MF_BYCOMMAND | MF_STRING, IDBTN_REW, "Reverse skip");
                  InsertMenu(hMenuMain, IDBTN_EXIT, MF_BYCOMMAND | MF_STRING, IDBTN_FF, "Forward skip");
                  InsertMenu(hMenuMain, IDBTN_EXIT, MF_BYCOMMAND | MF_SEPARATOR, NULL, NULL);
                  // add the taskbar icon
                  TaskBarAddIcon("FreeCD");
                  break;
               }
               else // if currently minimized...
               {
               	TaskBarDeleteIcon(); // remove the taskbar icon
                  // change the menu for use with the main window...
                  InsertMenu(hMenuMain, IDBTN_MIN, MF_BYCOMMAND | MF_STRING, IDBTN_ONTOP, "Always on top");
                  if (TopMost)
                  	CheckMenuItem(hMenuMain, IDBTN_ONTOP, MF_BYCOMMAND | MF_CHECKED);
                  ModifyMenu(hMenuMain, IDBTN_MIN, MF_BYCOMMAND | MF_STRING, IDBTN_MIN, "Minimize to taskbar");
                  DeleteMenu(hMenuMain, IDBTN_EJ, MF_BYCOMMAND);
                  DeleteMenu(hMenuMain, IDBTN_RPT, MF_BYCOMMAND);
                  DeleteMenu(hMenuMain, IDBTN_REW, MF_BYCOMMAND);
                  DeleteMenu(hMenuMain, IDBTN_FF, MF_BYCOMMAND);
                  DeleteMenu(hMenuMain, 5, MF_BYPOSITION);
                  // show the main window
                  ShowWindow(hWndMain, SW_SHOW);
                  SetForegroundWindow(hWndMain); // give the main window the input focus
                  Hidden = FALSE;
               }
               break;

            case IDBTN_EJ: // eject...
            	mciSendString((LPSTR)"status cdaudio mode", Buffer, 5, NULL);
               if (!strncmp(Buffer, "open", 4)) // if the drive is open, close it
               	mciSendString((LPSTR)"set cdaudio door closed", NULL, NULL, NULL);
               else // otherwise, open it
               {
               	mciSendString((LPSTR)"set cdaudio door open", NULL, NULL, NULL);
                  Status = open;
                  Repeat = FALSE;
                  UpdatePlayBtn();
               }
               // IMPORTANT NOTE: if the drive is closed with no disc in the
               // drive, the "status cdaudio mode" command will still return
               // "open", so the eject button won't be able to be used to open
               // the drive; I haven't been able to find any way around this
               // yet, even using different commands and variations...
               break;

            case IDBTN_ONTOP: // always-on-top...
            	if(TopMost) // if a-o-t is enabled...
               {
               	// turn off a-o-t
               	SetWindowPos(hWndMain, HWND_NOTOPMOST, 0,0,0,0, SWP_NOMOVE | SWP_NOSIZE);
                  TopMost = FALSE;
                  // uncheck the a-o-t menu item
                  CheckMenuItem(hMenuMain, IDBTN_ONTOP, MF_BYCOMMAND | MF_UNCHECKED);
               }
               else // if a-o-t is disabled...
               {
               	// turn on a-o-t
               	SetWindowPos(hWndMain, HWND_TOPMOST, 0,0,0,0, SWP_NOMOVE | SWP_NOSIZE);
                  TopMost = TRUE;
                  // check the a-o-t menu item
                  CheckMenuItem(hMenuMain, IDBTN_ONTOP, MF_BYCOMMAND | MF_CHECKED);
               }
               break;

            default:
            	break;
         }
         return DefWindowProc(hWndMain, message, wParam, lParam);

      case WM_LBUTTONDBLCLK: // the left mouse button was double-clicked...
      	// get the current mouse coordinates...
      	newPt.x = LOWORD(lParam);
         newPt.y = HIWORD(lParam);
         // if the mouse position is within the display rectangle...
         if (PtInRect(&rcDisp, newPt))
         {
				// change the display mode to the next mode...
         	switch (Disp)
            {
            	case norm:
               	Disp = remtrk;
                  break;
               case remtrk:
               	Disp = remdisc;
                  break;
               case remdisc:
               	Disp = total;
                  break;
               case total:
               	Disp = norm;
                  break;
            }
            break;
         }
         return DefWindowProc(hWndMain, message, wParam, lParam);

      case WM_LBUTTONDOWN: // the left mouse button was single-clicked...
         // get the mouse position...
      	oldPt.x = LOWORD(lParam);
         oldPt.y = HIWORD(lParam);
         // if the position is inthe display rectangle...
         if (PtInRect(&rcDisp, oldPt))
         {
         	// capture the mouse input
         	SetCapture(hWndMain);
            BtnDn = TRUE; // we're now in DRAG mode
            break;
         }
         return DefWindowProc(hWndMain, message, wParam, lParam);

      case WM_MOUSEMOVE: // the mouse was moved...
      	if (BtnDn) // if we're in DRAG mode (the mouse is moved while the button is held down)...
         {
         	// get the new mouse position...
         	newPt.x = LOWORD(lParam);
            newPt.y = HIWORD(lParam);
            // adjust the window position to reflect the mouse movement...
            WinX = WinX - (oldPt.x - newPt.x);
            WinY = WinY - (oldPt.y - newPt.y);
            // move the window to the new position
            MoveWindow(hWndMain, WinX, WinY, WinWidth, WinHeight, TRUE);
         }
         return DefWindowProc(hWndMain, message, wParam, lParam);

      case WM_LBUTTONUP: // the left mouse button was released...
      	if (BtnDn) // if we're DRAGGING...
         {
         	RECT rc;
            ReleaseCapture(); // release the mouse capture
            BtnDn = FALSE; // no longer in DRAG mode
            GetWindowRect(hWndMain, &rc); // get the coords of the main window
            // and assign them to the x & y global window position variables
            WinX = rc.left;
            WinY = rc.top;
            break;
         }
         return DefWindowProc(hWndMain, message, wParam, lParam);

      case CDWM_NOTIFYICON: // the taskbar icon has received input...
      	On_CDWM_NOTIFYICON(lParam); // call the handler procedure
         break;

      case WM_CONTEXTMENU: // the right mouse button was clicked in the main window...
      	// open the context menu; if it fails, pass the message off
         // to the default message handler
      	if (!OnContextMenu(LOWORD(lParam), HIWORD(lParam)))
         	return DefWindowProc(hWndMain, message, wParam, lParam);
         break;

      case WM_CREATE: // the main window is being created...
      	HMENU hMenu;

         PrimeBmpOps(); // prepare the window-refresh bitmap
         hMenu = LoadMenu(hInst, "MainMenu"); // open the menu resource
         hMenuMain = GetSubMenu(hMenu, 0); // get the first popup menu in that resource

			// insert an owner-drawn menu item
         InsertMenu(hMenuMain, IDBTN_OPT, MF_BYCOMMAND | MF_OWNERDRAW, IDBMP_LOGO, NULL);

			// get the default system menu, and change it...
         hMenuSys = GetSystemMenu(hWndMain, FALSE);
         ModifyMenu(hMenuSys, 0, MF_BYPOSITION | MF_STRING, IDBTN_HELP, "Help");
         ModifyMenu(hMenuSys, 1, MF_BYPOSITION | MF_STRING, IDBTN_MIN, "Minimize");
         DeleteMenu(hMenuSys, SC_SIZE, MF_BYCOMMAND);
         DeleteMenu(hMenuSys, SC_MINIMIZE, MF_BYCOMMAND);
         DeleteMenu(hMenuSys, SC_MAXIMIZE, MF_BYCOMMAND);

			// load the taskbar icons...
         hIconPlay = LoadIcon(hInst, "TBPlay");
         hIconPause = LoadIcon(hInst, "TBPause");
         hIconX = LoadIcon(hInst, "TBX");

         // create the buttons...
         CreateWindow("button", NULL, WS_CHILD | WS_VISIBLE | BS_OWNERDRAW,
         	ref[REW].x, ref[REW].y, 29, 30, hWndMain, (HMENU)IDBTN_REW, hInst, NULL);
         CreateWindow("button", NULL, WS_CHILD | WS_VISIBLE | BS_OWNERDRAW,
         	ref[PLAY].x, ref[PLAY].y, 79, 30, hWndMain, (HMENU)IDBTN_PLAY, hInst, NULL);
         CreateWindow("button", NULL, WS_CHILD | WS_VISIBLE | BS_OWNERDRAW,
         	ref[FF].x, ref[FF].y, 29, 30, hWndMain, (HMENU)IDBTN_FF, hInst, NULL);
         CreateWindow("button", NULL, WS_CHILD | WS_VISIBLE | BS_OWNERDRAW,
         	ref[VOL].x, ref[VOL].y, 15, 15, hWndMain, (HMENU)IDBTN_VOL, hInst, NULL);
         CreateWindow("button", NULL, WS_CHILD | WS_VISIBLE | BS_OWNERDRAW,
         	ref[RPT].x, ref[RPT].y, 15, 15, hWndMain, (HMENU)IDBTN_RPT, hInst, NULL);
         CreateWindow("button", NULL, WS_CHILD | WS_VISIBLE | BS_OWNERDRAW,
         	ref[MIN].x, ref[MIN].y, 15, 15, hWndMain, (HMENU)IDBTN_MIN, hInst, NULL);
         CreateWindow("button", NULL, WS_CHILD | WS_VISIBLE | BS_OWNERDRAW,
         	ref[EJ].x, ref[EJ].y, 15, 15, hWndMain, (HMENU)IDBTN_EJ, hInst, NULL);
         CreateWindow("button", "&x", WS_CHILD | WS_VISIBLE | BS_OWNERDRAW,
         	ref[EXIT].x, ref[EXIT].y, 15, 15, hWndMain, (HMENU)IDBTN_EXIT, hInst, NULL);
         CreateWindow("button", "&h", WS_CHILD | WS_VISIBLE | BS_OWNERDRAW,
         	ref[HELP].x, ref[HELP].y, 15, 15, hWndMain, (HMENU)IDBTN_HELP, hInst, NULL);

         // this is unorthodox, but:
         // post a message to carry out further preparations
         PostMessage(hWndMain, CDWM_NEWWIN, NULL, NULL);

         return (DefWindowProc(hWndMain, message, wParam, lParam));

      case CDWM_NEWWIN: // carry out further preparations...
      	TIMECAPS tc;

			// open the cdaudio device as shareable with other programs
			mciSendString((LPSTR)"open cdaudio shareable", NULL, NULL, NULL);
         mciSendString((LPSTR)"set cdaudio time format tmsf", NULL, NULL, NULL);

			// is there a disc in the drive?
         mciSendString((LPSTR)"status cdaudio media present", Buffer, 5, NULL);
         if (!strncmp(Buffer, "true", 4)) // if yes...
         {
         	DiscPres = TRUE;
            mciSendString((LPSTR)"status cdaudio mode", Buffer, 8, NULL);
            if (strncmp(Buffer, "playing", 7)) // if it isn't playing...
         	{
			      if (Opt.AutoPlay) // play it automatically?
	      		{
               	PlayDisc();
		   	   	Status = playing;
				      UpdatePlayBtn();
               }
            }
            else // if it is playing...
            {
		   		Status = playing;
					UpdatePlayBtn();
            }
		      strcpy(ref[DISP].bmpup, "BmpClrTime");
		      DrawBmp(ref[DISP].x, ref[DISP].y, ref[DISP].bmpup); // clear the time display
		      GetDiscInfo();
		      // make sure the disc length is displayed for
		      // at least a couple seconds...
		      Wait = 2;
         }
         timeGetDevCaps(&tc, sizeof(TIMECAPS)); // get timer device capabilities
         wTimerRes = min(max(tc.wPeriodMin, 1), tc.wPeriodMax); // get optimum timer resolution
         timeBeginPeriod(wTimerRes); // set timer resolution
			// start the timer
         wTimerID = timeSetEvent(TimeOut, wTimerRes, (LPTIMECALLBACK)TimerProc, NULL, TIME_PERIODIC);
         break;

		case WM_SHOWWINDOW: // the window is about to be shown or hidden...
      	BOOL Show;
         Show = (BOOL)wParam;
         if (Show) // if it's about to be shown...
         {
         	RECT rc, rcWin;
            APPBARDATA abd;

            abd.cbSize = sizeof(APPBARDATA);
            abd.hWnd = hWndMain;

            GetWindowRect(GetDesktopWindow(), &rcWin); // get the size of the screen
            SHAppBarMessage(ABM_GETTASKBARPOS, &abd); // get the size & position of the taskbar
            // subtract the taskbar rectangle from the screen rectangle
            // to get the usable screen area
            SubtractRect(&rc, &rcWin, &abd.rc);

				// adjust the window position so that it doesn't
            // overlap the taskbar...
            if (WinX < rc.left)
            	WinX = rc.left;
            if (WinY < rc.top)
            	WinY = rc.top;
            if ( (WinX + WinWidth) > rc.right )
            	WinX = rc.right - WinWidth;
            if ( (WinY + WinHeight) > rc.bottom )
            	WinY = rc.bottom - WinHeight;

            // move the window to the new position
            SetWindowPos(hWndMain, NULL, WinX, WinY, NULL, NULL,
            	SWP_NOSIZE | SWP_NOZORDER);
            return 0;
         }
         return (DefWindowProc(hWndMain, message, wParam, lParam));

      case WM_PAINT: // the window needs to be redrawn...
      	if (!Hidden) // if it's not minimized...
         {
            HDC hdcMain;
            BeginPaint(hWndMain, &ps);
            hdcMain = GetDC(hWndMain);
            // copy the window-refresh bitmap from memory to the screen
            BitBlt(hdcMain, 0, 0, WinWidth, WinHeight, hdcMemMain, 0, 0,  SRCCOPY);
            ReleaseDC(hWndMain, hdcMain);
            EndPaint(hWndMain, &ps);
         }
         return 0;

      case WM_DESTROY: // the window is being detroyed...
      	PostQuitMessage(0);
         break;

      case WM_CLOSE: // FreeCD should close...
      	KillApp();
         break;

      case WM_ENDSESSION: // the Windows session might be ending...
      	if ((BOOL)wParam) // yep, it's ending...
         	KillApp();
         return 0;

      default: // pass the message on to the default message handler
      	return(DefWindowProc(hWndMain, message, wParam, lParam));
	}
	return (0L);
}
